@@ -16,6 +16,7 @@ package web
 import (
 	"bytes"
 	"context"
+	"errors"
 	"fmt"
 	"io"
 	"net/http"
@@ -25,7 +26,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
 	"github.com/prometheus/common/model"
 	"github.com/stretchr/testify/require"
 
@@ -37,6 +37,7 @@ import (
 	"github.com/prometheus/prometheus/storage"
 	"github.com/prometheus/prometheus/tsdb"
 	"github.com/prometheus/prometheus/util/teststorage"
+	"github.com/prometheus/prometheus/util/testutil"
 )
 
 var scenarios = map[string]struct {
@@ -47,29 +48,29 @@ var scenarios = map[string]struct {
 }{
 	"empty": {
 		params: "",
-		code:   200,
+		code:   http.StatusOK,
 		body:   ``,
 	},
 	"match nothing": {
 		params: "match[]=does_not_match_anything",
-		code:   200,
+		code:   http.StatusOK,
 		body:   ``,
 	},
 	"invalid params from the beginning": {
 		params: "match[]=-not-a-valid-metric-name",
-		code:   400,
+		code:   http.StatusBadRequest,
 		body: `1:1: parse error: unexpected <op:->
 `,
 	},
 	"invalid params somewhere in the middle": {
 		params: "match[]=not-a-valid-metric-name",
-		code:   400,
+		code:   http.StatusBadRequest,
 		body: `1:4: parse error: unexpected <op:->
 `,
 	},
 	"test_metric1": {
 		params: "match[]=test_metric1",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -77,33 +78,33 @@ test_metric1{foo="boo",instance="i"} 1 6000000
 	},
 	"test_metric2": {
 		params: "match[]=test_metric2",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric2 untyped
 test_metric2{foo="boo",instance="i"} 1 6000000
 `,
 	},
 	"test_metric_without_labels": {
 		params: "match[]=test_metric_without_labels",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric_without_labels untyped
 test_metric_without_labels{instance=""} 1001 6000000
 `,
 	},
 	"test_stale_metric": {
 		params: "match[]=test_metric_stale",
-		code:   200,
+		code:   http.StatusOK,
 		body:   ``,
 	},
 	"test_old_metric": {
 		params: "match[]=test_metric_old",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric_old untyped
 test_metric_old{instance=""} 981 5880000
 `,
 	},
 	"{foo='boo'}": {
 		params: "match[]={foo='boo'}",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="boo",instance="i"} 1 6000000
 # TYPE test_metric2 untyped
@@ -112,7 +113,7 @@ test_metric2{foo="boo",instance="i"} 1 6000000
 	},
 	"two matchers": {
 		params: "match[]=test_metric1&match[]=test_metric2",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -122,7 +123,7 @@ test_metric2{foo="boo",instance="i"} 1 6000000
 	},
 	"two matchers with overlap": {
 		params: "match[]={__name__=~'test_metric1'}&match[]={foo='bar'}",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -130,7 +131,7 @@ test_metric1{foo="boo",instance="i"} 1 6000000
 	},
 	"everything": {
 		params: "match[]={__name__=~'.%2b'}", // '%2b' is an URL-encoded '+'.
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -144,7 +145,7 @@ test_metric_without_labels{instance=""} 1001 6000000
 	},
 	"empty label value matches everything that doesn't have that label": {
 		params: "match[]={foo='',__name__=~'.%2b'}",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric_old untyped
 test_metric_old{instance=""} 981 5880000
 # TYPE test_metric_without_labels untyped
@@ -153,7 +154,7 @@ test_metric_without_labels{instance=""} 1001 6000000
 	},
 	"empty label value for a label that doesn't exist at all, matches everything": {
 		params: "match[]={bar='',__name__=~'.%2b'}",
-		code:   200,
+		code:   http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -168,7 +169,7 @@ test_metric_without_labels{instance=""} 1001 6000000
 	"external labels are added if not already present": {
 		params:         "match[]={__name__=~'.%2b'}", // '%2b' is an URL-encoded '+'.
 		externalLabels: labels.FromStrings("foo", "baz", "zone", "ie"),
-		code:           200,
+		code:           http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i",zone="ie"} 10000 6000000
 test_metric1{foo="boo",instance="i",zone="ie"} 1 6000000
@@ -185,7 +186,7 @@ test_metric_without_labels{foo="baz",instance="",zone="ie"} 1001 6000000
 		// know what it does anyway.
 		params:         "match[]={__name__=~'.%2b'}", // '%2b' is an URL-encoded '+'.
 		externalLabels: labels.FromStrings("instance", "baz"),
-		code:           200,
+		code:           http.StatusOK,
 		body: `# TYPE test_metric1 untyped
 test_metric1{foo="bar",instance="i"} 10000 6000000
 test_metric1{foo="boo",instance="i"} 1 6000000
@@ -223,7 +224,7 @@ func TestFederation(t *testing.T) {
 	for name, scenario := range scenarios {
 		t.Run(name, func(t *testing.T) {
 			h.config.GlobalConfig.ExternalLabels = scenario.externalLabels
-			req := httptest.NewRequest("GET", "http://example.org/federate?"+scenario.params, nil)
+			req := httptest.NewRequest(http.MethodGet, "http://example.org/federate?"+scenario.params, nil)
 			res := httptest.NewRecorder()
 
 			h.federation(res, req)
@@ -238,15 +239,15 @@ type notReadyReadStorage struct {
 }
 
 func (notReadyReadStorage) Querier(int64, int64) (storage.Querier, error) {
-	return nil, errors.Wrap(tsdb.ErrNotReady, "wrap")
+	return nil, fmt.Errorf("wrap: %w", tsdb.ErrNotReady)
 }
 
 func (notReadyReadStorage) StartTime() (int64, error) {
-	return 0, errors.Wrap(tsdb.ErrNotReady, "wrap")
+	return 0, fmt.Errorf("wrap: %w", tsdb.ErrNotReady)
 }
 
 func (notReadyReadStorage) Stats(string, int) (*tsdb.Stats, error) {
-	return nil, errors.Wrap(tsdb.ErrNotReady, "wrap")
+	return nil, fmt.Errorf("wrap: %w", tsdb.ErrNotReady)
 }
 
 // Regression test for https://github.com/prometheus/prometheus/issues/7181.
@@ -264,7 +265,7 @@ func TestFederation_NotReady(t *testing.T) {
 				},
 			}
 
-			req := httptest.NewRequest("GET", "http://example.org/federate?"+scenario.params, nil)
+			req := httptest.NewRequest(http.MethodGet, "http://example.org/federate?"+scenario.params, nil)
 			res := httptest.NewRecorder()
 
 			h.federation(res, req)
@@ -354,7 +355,7 @@ func TestFederationWithNativeHistograms(t *testing.T) {
 			_, err = app.AppendHistogram(0, l, 100*60*1000, histWithoutZeroBucket.Copy(), nil)
 			expVec = append(expVec, promql.Sample{
 				T:      100 * 60 * 1000,
-				H:      histWithoutZeroBucket.ToFloat(),
+				H:      histWithoutZeroBucket.ToFloat(nil),
 				Metric: expL,
 			})
 		default:
@@ -363,7 +364,7 @@ func TestFederationWithNativeHistograms(t *testing.T) {
 			_, err = app.AppendHistogram(0, l, 100*60*1000, hist.Copy(), nil)
 			expVec = append(expVec, promql.Sample{
 				T:      100 * 60 * 1000,
-				H:      hist.ToFloat(),
+				H:      hist.ToFloat(nil),
 				Metric: expL,
 			})
 		}
@@ -380,7 +381,7 @@ func TestFederationWithNativeHistograms(t *testing.T) {
 		},
 	}
 
-	req := httptest.NewRequest("GET", "http://example.org/federate?match[]=test_metric", nil)
+	req := httptest.NewRequest(http.MethodGet, "http://example.org/federate?match[]=test_metric", nil)
 	req.Header.Add("Accept", `application/vnd.google.protobuf;proto=io.prometheus.client.MetricFamily;encoding=delimited,application/openmetrics-text;version=1.0.0;q=0.8,application/openmetrics-text;version=0.0.1;q=0.75,text/plain;version=0.0.4;q=0.5,*/*;q=0.1`)
 	res := httptest.NewRecorder()
 
@@ -389,14 +390,13 @@ func TestFederationWithNativeHistograms(t *testing.T) {
 	require.Equal(t, http.StatusOK, res.Code)
 	body, err := io.ReadAll(res.Body)
 	require.NoError(t, err)
-
-	p := textparse.NewProtobufParser(body, false)
+	p := textparse.NewProtobufParser(body, false, labels.NewSymbolTable())
 	var actVec promql.Vector
 	metricFamilies := 0
 	l := labels.Labels{}
 	for {
 		et, err := p.Next()
-		if err == io.EOF {
+		if err != nil && errors.Is(err, io.EOF) {
 			break
 		}
 		require.NoError(t, err)
@@ -427,5 +427,5 @@ func TestFederationWithNativeHistograms(t *testing.T) {
 	// TODO(codesome): Once PromQL is able to set the CounterResetHint on histograms,
 	// test it with switching histogram types for metric families.
 	require.Equal(t, 4, metricFamilies)
-	require.Equal(t, expVec, actVec)
+	testutil.RequireEqual(t, expVec, actVec)
 }
